/*
 * Copyright (c) 2014, Artem Boldarev <artem.boldarev@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <curses.h>

/*   Windows */
#ifdef WIN32

#undef MOUSE_MOVED /*   curses.h,   windows.h */
#include <windows.h>
#include <ufmod.h>

/*       -     */
#define PLAY_SONG() (uFMOD_PlaySong((void *)music, (void *)sizeof(music), XM_MEMORY))
#define STOP_SONG() (uFMOD_PlaySong(NULL, NULL, 0))

/*   Linux */
#elif __linux__

#include <unistd.h>
#include <alsa/asoundlib.h>
#include <alsaufmod.h>

#define PLAY_SONG() (uFMOD_ALSAPlaySong((void *)music, sizeof(music), XM_MEMORY, "plughw:0,0"))
#define STOP_SONG() (uFMOD_ALSAPlaySong(NULL, (int) NULL, 0, NULL))

#else
#error Unsupported OS!
#endif

/*   */
#include "animation.h" /* Stolen from: https://github.com/klange/nyancat/ */
/*  */
#include "music.h" /* Original XM found here: https://github.com/djui/nyancat/ */

#define ESC (27) /*   Escape */

/*
    curses  .
       ( - ).
*/
#define TICKRATE (50) 

void draw_frame(char *frame[]);

int main(int argc, char *argv[])
{
	int c;
	unsigned int fi = 0; /*    */

	/*   */
	initscr();
	noecho();
	cbreak();
	keypad(stdscr, TRUE);
	curs_set(FALSE); /*   */
	/*   */
	timeout(TICKRATE);

	/*   */
	if (has_colors())
	{
		start_color();

		/*    */
		init_pair(COLOR_BLACK, COLOR_BLACK, COLOR_BLACK);
		init_pair(COLOR_GREEN, COLOR_GREEN, COLOR_BLACK);
		init_pair(COLOR_RED, COLOR_RED, COLOR_BLACK);
		init_pair(COLOR_CYAN, COLOR_CYAN, COLOR_BLACK);
		init_pair(COLOR_WHITE, COLOR_WHITE, COLOR_BLACK);
		init_pair(COLOR_MAGENTA, COLOR_MAGENTA, COLOR_BLACK);
		init_pair(COLOR_BLUE, COLOR_BLUE, COLOR_BLACK);
		init_pair(COLOR_YELLOW, COLOR_YELLOW, COLOR_BLACK);
	}

#ifdef PDCURSES
	/*    ,    */
	if (is_termresized())
		resize_term(FRAME_HEIGHT, FRAME_WIDTH);

	/*    */
	PDC_set_title("Nyanyanyanya-nya-nya-nya-nya-nya...");
#endif

	/*   */
	PLAY_SONG();

	/*    */
	while ((c = getch()) != ESC) /*    Escape */
	{
		if (c == ERR) /*  TICKRATE  -   */
		{
			/*   */
#ifdef WIN32
			clear(); /*  Windows    ,     */
#endif
			draw_frame(frames[fi]);
			/*   */
			fi++;
			if (frames[fi] == NULL) /*   */
				fi = 0; /*    */
			refresh();
		}
	}

	/*    */
	endwin();
	/*   */
	STOP_SONG();

	return 0;
}

/*
   .
,  FRAME_WIDTH  FRAME_HEIGHT
, ,    .
*/
void draw_frame(char *frame[])
{
	int lines_offset;
	int width_correction;
	int i, j;

	move(0, 0); /*    */
	/*      */
	attron(COLOR_PAIR(COLOR_BLUE));
	for (i = 0; i < COLS; i++)
	{
		for (j = 0; j < LINES; j++)
		{
			addch(ACS_BLOCK);
		}
	}
	attroff(COLOR_PAIR(COLOR_BLUE));

	move(0, 0); /*    */
	/* !!!!   !!!! */
	/*
	    ,   .
	   ,    .
 	*/
	lines_offset = (FRAME_HEIGHT - LINES) / 2;
	if (lines_offset < 0)
		lines_offset = 0;
	/*
	,         ,
	         .
	*/
	width_correction = (COLS < FRAME_WIDTH ? FRAME_WIDTH - COLS + 1 : 0);
	/*   */
	for (i = lines_offset; i < LINES + lines_offset && i < FRAME_HEIGHT; i++)
	{
		int y;
		for (j = 0; j < COLS && j < (FRAME_WIDTH - width_correction); j++)
		{
			chtype a; /*    */
			chtype oa; /*    */

			/*    */
			oa = a = getattrs(stdscr);
			/* !!!     !!! */
			switch (frame[i][j])
			{
				/*    */
				case ',':
					a = COLOR_PAIR(COLOR_BLUE);
					break;
				/*   */
				case '.':
					a = COLOR_PAIR(COLOR_WHITE) | A_BOLD;
					break;
				/* ׸  */
				case '\'':
					a = COLOR_PAIR(COLOR_BLACK);
					break;
				/*     */
				case '@':
					a = COLOR_PAIR(COLOR_WHITE) | A_BOLD;
					break;
				/* -    */
				case '$':
					a = COLOR_PAIR(COLOR_MAGENTA) | A_BOLD;
					break;
				/*     */
				case '-':
					a = COLOR_PAIR(COLOR_MAGENTA);
					break;
				/*     */
				case '>':
					a = COLOR_PAIR(COLOR_RED);
					break;
				/*     */
				case '&':
					a = COLOR_PAIR(COLOR_RED) | A_BOLD;
					break;
				/* Ƹ    */
				case '+':
					a = COLOR_PAIR(COLOR_YELLOW) | A_BOLD;
					break;
				/*     */
				case '#':
					a = COLOR_PAIR(COLOR_GREEN) | A_BOLD;
					break;
				/* -    */
				case '=':
					a = COLOR_PAIR(COLOR_CYAN) | A_BOLD;
					break;
				/* -    */
				case ';':
					a = COLOR_PAIR(COLOR_BLUE) | A_BOLD;
					break;
				/*    */
				case '*':
					a = COLOR_PAIR(COLOR_WHITE);
					break;
				/*      */
				case '%':
					a = COLOR_PAIR(COLOR_MAGENTA) | A_BOLD;
					break;
			}

			/*    -   */
			if ((a != oa))
			{
				attron(a);
				addch(ACS_BLOCK);
				attroff(a);
			}
			else /* ,     */
			{
				attron(COLOR_PAIR(COLOR_BLACK));
				addch(' ');
				attroff(COLOR_PAIR(COLOR_BLACK));
			}

		}

		/*       */
		y = getcury(stdscr);
		move(y + 1, 0);
	}
}

