/*
 * Copyright (c) 2014, Artem Boldarev <artem.boldarev@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#define BYTES_PER_LINE (8)

#define SETBIT(v, n) ((v) |= (1<<(n)))
#define IS_SET(v, n) ((v) & (1<<(n)))

#define PRINT_HEX " 0x%.2x"
#define PRINT_OCT " 0%.3o"
#define PRINT_DEC " %d"

/* must be power of two */
enum array_type {
	ARR_STATIC = 1,
	ARR_CONST = 2
};

typedef enum _data_format_t {
	FMT_HEX = 0,
	FMT_OCT,
	FMT_DEC
} data_format_t;

/* Program configuration */
struct _cfg {
	unsigned long array_type;
	data_format_t data_fmt;	/* hex, oct, dec */
	char *arr_name;
	int use_include_guard;	/* generate header protected by include guard */
	char *in_fname;
	FILE *fin;
	FILE *fout;
}
cfg;

/* full program name  */
char *argv0 = NULL;

int dumpfile(void);
long fsize(FILE *fd);
char *strtoid(char *s);
void usage(void);

int main(int argc, char *argv[])
{
	int i, argleft;
	int get_arr_name = 0;

	memset(&cfg, 0, sizeof(cfg));
	argv0 = argv[0];
	/* parse command line options */
	for (i = 1; i < argc; i++)
	{
		char *arg;
		arg = argv[i];

		if (get_arr_name)
		{
			cfg.arr_name = arg;
			get_arr_name = 0;
			continue;
		}

		if (arg[0] == '-' || arg[0] == '/')
		{
			char *p;
			if (arg[1] == '\0')	/* last option */
				break;


			for (p = ++arg; *p != '\0'; p++)
			{
				int argch = *p;
				switch (argch)
				{
					case 'g':
						cfg.use_include_guard = 1;
						break;
					case 'a':
						/* obtain array name */
						if (p[1] != '\0' && p[1] != '-' && p[1] != ' ')
							cfg.arr_name = &p[1];
						else
							get_arr_name = 1;
						break;
					case 'o':
						cfg.data_fmt = FMT_OCT;
						break;
					case 'd':
						cfg.data_fmt = FMT_DEC;
						break;
					case 's':
						SETBIT(cfg.array_type, ARR_STATIC);
						break;
					case 'c':
						SETBIT(cfg.array_type, ARR_CONST);
						break;
					case 'h':
					case '?':
						usage();
						break;
					default:
						fprintf(stderr, "Unknown option %c\n\n", argch);
						usage();
						break;
				}
			}
		}
		else
			break;
	}

	argleft = argc - i;
	argv += i;

	if (cfg.arr_name == NULL)
	{
		fprintf(stderr, "Array name is not specified!\n");
		usage();
	}

	if (argleft < 1)
	{
		fprintf(stderr, "Input file is not specified!\n");
		usage();
	}

	cfg.in_fname = argv[0];
	if ((cfg.fin = fopen(argv[0], "rb")) == NULL)
	{
		perror("Can\'t open input file!");
		return 1;
	}

	if (argleft >= 2)
	{
		if ((cfg.fout = fopen(argv[1], "w")) == NULL)
		{
			perror("Can\'t open output file!");
			return 1;
		}

	}
	else
		cfg.fout = stdout;

	if (dumpfile() != 0)
		perror("ERROR");

	fclose(cfg.fin);
	if (cfg.fout != stdout)
		fclose(cfg.fout);

	return 0;
}

void usage(void)
{
	/* obtain executable name */
	char *exe_name;
	if ((exe_name = strrchr(argv0, '\\')) == NULL && (exe_name = strrchr(argv0, '/')) == NULL)
		exe_name = argv0;
	else
		exe_name++;

	fprintf(stderr, "\nGenerates C/C++ header from arbitrary file.\n");
	fprintf(stderr, "Copyright (c) 2013, Artem Boldarev <artem.boldarev@gmail.com>\n");

	fprintf(stderr, "Usage:\n" "%s -[godsch] -a <array_name> <input_file> <output_file (optional)>\n", exe_name);
	fprintf(stderr, "Options:\n"
		" a <array name> - set array name inside the generated header file\n"
		" g - generate header file protected with \'include guard\'\n"
		" o - save data as octal numbers\n"
		" d - save data as decimal numbers\n"
		" s - \'static\' array\n"
		" c - \'const\' array\n"
		" h - show this help and exit\n"
		"If output file is not set - standard output used instead.\n");

	exit(1);
}

char *strtoid(char *s)
{
	char *p;

	p = s;
	while (*p != '\0')
	{
		if (isdigit(*p) || isalpha(*p))
			*p = toupper(*p);
		else
			*p = '_';

		p++;
	}

	return s;
}

long fsize(FILE *fd)
{
	long old_offset;
	long size;

	if ((old_offset = ftell(fd)) == -1)
		return -1;
	if (fseek(fd, 0, SEEK_END) != 0)
	{
		fseek(fd, old_offset, SEEK_SET);
		return -1;
	}

	size = ftell(fd);

	if (fseek(fd, old_offset, SEEK_SET) != 0)
		return -1;

	return size;
}

/*
 * There is so many error handling in that function because
 * this program created to be included in other projects controlled
 * by make utility. So if we have any error we should fail as fast as possible
 * to disallow generate incorrect results.
 */
int dumpfile(void)
{
	char *byte_fmt;
	char guard_id[100];
	char tbuf[100] = { 0 };
	time_t t;
	struct tm *lt;
	long file_size;
	unsigned long offset;
	size_t readen;

	/* set output format */
	switch (cfg.data_fmt)
	{
		case FMT_HEX:
			byte_fmt = PRINT_HEX;
			break;
		case FMT_OCT:
			byte_fmt = PRINT_OCT;
			break;
		case FMT_DEC:
			byte_fmt = PRINT_DEC;
			break;
		default:
			byte_fmt = PRINT_HEX;
			break;
	}

	/* obtain file size */
	file_size = fsize(cfg.fin);
	if (file_size == -1)
		return -1;

	/* get generation time */
	t = time(NULL);
	lt = localtime(&t);
	if (lt == NULL)
		return 1;
	strftime(tbuf, 100, "%c", lt);

	if (fprintf(cfg.fout, "/* THIS FILE IS AUTOMATICALLY GENERATED!!! DO NOT EDIT!!! */\n") < 0)
		return 1;

	if (fprintf(cfg.fout, "/* Generated on: [ %s ], from file: %s */\n\n", tbuf, cfg.in_fname) < 0)
		return 1;

	if (cfg.use_include_guard)
	{
		snprintf(guard_id, 100, "_%s_H", cfg.arr_name);
		strtoid(guard_id);

		if (fprintf(cfg.fout, "#ifndef %s\n#define %s\n\n", guard_id, guard_id) < 0)
			return 1;
	}

	if (fprintf(cfg.fout, 
				"%s%sunsigned char %s[] = {\n",
			       	(IS_SET(cfg.array_type, ARR_STATIC) ? "static " : ""),
			       	(IS_SET(cfg.array_type, ARR_CONST) ? "const " : ""), cfg.arr_name)
		       	< 0)
	{
		return 1;
	}

	for (offset = readen = 0; offset != file_size && !feof(cfg.fin); offset += readen)
	{
		int i = 0;
		unsigned char buf[BYTES_PER_LINE];

		/* padding */
		if (fprintf(cfg.fout, "/* 0x%.8lx: */ ", offset) < 0)
			return 1;

		/* print  offset */
		for (readen = 0; readen != BYTES_PER_LINE && fread(&buf[readen], 1, 1, cfg.fin) > 0; readen++)
			;


		if (ferror(cfg.fin))
			return 1;

		for (i = 0; i < readen; i++)
		{
			if (fprintf(cfg.fout, byte_fmt, (unsigned char)buf[i]) < 0)
				return 1;

			if ((offset + i) != (file_size - 1))
			{
				if (fprintf(cfg.fout, ",") < 0)
					return 1;
			}

		}

		if (fprintf(cfg.fout, "\n") < 0)
			return 1;
	}

	if (fprintf(cfg.fout, "};\n") < 0)
		return 1;

	if (cfg.use_include_guard)
	{
		if (fprintf(cfg.fout, "\n#endif /* %s */\n", guard_id) < 0)
			return 1;
	}

	return 0;
}

